
#ifndef HAVE_BONDED_PAIRS_HH
#define HAVE_BONDED_PAIRS_HH

#include <string>
#include <vector>
#include <iostream>

#include <mmdb/mmdb_manager.h>

#include "protein-geometry.hh"

namespace coot {

   // The residues here are in order.  res_1 is comp_1 and res_2 is comp_2
   class bonded_pair_t {
      void delete_atom(CResidue *res, const std::string &atom_name);
   public:
      CResidue *res_1;
      CResidue *res_2;
      std::string link_type;
      bool is_fixed_first;
      bool is_fixed_second;
      bonded_pair_t(CResidue *r1, CResidue *r2, bool is_fixed_first_in, bool is_fixed_second_in,
		    const std::string &lt) {
	 res_1 = r1;
	 res_2 = r2;
	 link_type = lt;
	 is_fixed_first = is_fixed_first_in;
	 is_fixed_second = is_fixed_second_in;
      }
      bool matches(CResidue *r1, CResidue *r2) const {
	 if (r1 == res_1 && r2 == res_2) {
	    return true;
	 } else{
	    if (r1 == res_2 && r2 == res_1) {
	       return true;
	    } else {
	       return false;
	    } 
	 }
      }
      void apply_chem_mods(const protein_geometry &geom);
   };
   std::ostream &operator<<(std::ostream &s, bonded_pair_t bp);
   
   class bonded_pair_container_t {
   public:
      std::vector<bonded_pair_t> bonded_residues;
      bool try_add(const bonded_pair_t &bp); // check for null residues too.
      unsigned int size() const { return bonded_residues.size(); }
      bonded_pair_t operator[](unsigned int i) { return bonded_residues[i]; }
      const bonded_pair_t operator[](unsigned int i) const { return bonded_residues[i]; }
      bool linked_already_p(CResidue *r1, CResidue *r2) const;
      friend std::ostream& operator<<(std::ostream &s, bonded_pair_container_t bpc);
      // test order switch too.
      bool matches(CResidue *r1, CResidue *r2) const {
	 bool r = false;
	 for (unsigned int i=0; i<bonded_residues.size(); i++) { 
	    if (bonded_residues[i].matches(r1, r2)) {
	       r = true;
	       break;
	    } 
	 }
	 return r;
      } 
      void apply_chem_mods(const protein_geometry &geom);
   };
   std::ostream& operator<<(std::ostream &s, bonded_pair_container_t bpc);

}

#endif // HAVE_BONDED_PAIRS_HH
